do -- Section [1]: Loading process 
	FlexBar2 = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceModuleCore-2.0", "AceDebug-2.0", "AceEvent-2.0")
	FlexBar2.OO = AceLibrary("AceOO-2.0");
	FlexBar2.Waterfall = AceLibrary("Waterfall-1.0");

	FlexBar2:RegisterDB("FlexBar2_DB");
	FlexBar2:RegisterDefaults('char', {
		DataVersion = 1, -- Changed when incompatibilities occur in SavedVars during versions
		Buttons = {},
        Groups = {},
	});

	function FlexBar2:OnInitialize()
		-- This is set to true when i'm coding, false when I'm commiting :)
	    self:SetDebugging(false);

		-- Code to handle deprecated settings and copy to new location
		local OldSettings = self.db.profile;		
		if(type(OldSettings.Buttons) == "table") then
			-- Copy over old settings (profile based) to the new location (char based)
			self.db.char.Buttons = OldSettings.Buttons;
			self.db.char.Groups = OldSettings.Groups;
			-- Now we're done copying, nil out all old stuff so it doesn't get saved again
			self.db.profile.Buttons = nil;
			self.db.profile.Groups = nil;
			self.db.profile.DataVersion = nil;
		end
		-- Create all the tables we need for intializing
		self.ButtonList = {};
		self.ButtonMixinList = {};
		self.GroupMixinList = {};
		self.ButtonActivatorList = {};
		self.ButtonDeactivatorList = {};
		self.GroupActivatorList = {};
		self.GroupDeactivatorList = {};
		self.ButtonWaterfallList = {};
		self.GroupWaterfallList = {};
		self.GeneralWaterfallList = {};
		self.ButtonSettingLoaderList = {};
		self.ButtonDefaulters = {};
		self.Buttons = {};
        self.Groups = {};
		self:Debug("Searching for modules");

		-- Loop through all modules and find if certain stuff has been declared, if they have, add them to the table.
		-- this includes ButtonMixin:Activate(), ButtonMixin:Deactivate(), ButtonMixin:LoadSettings(), ButtonMixin:SetDefaults(), GroupMixin:Activate(), ButtonWaterfall & Groupwaterfall
	    for _, Name in ipairs(self.ModuleList) do
            local Module = self.modules[Name];
			self:Debug("     Found Module: " .. Name);
			if(type(Module.ButtonMixin) == "table") then
				table.insert(self.ButtonMixinList, Module.ButtonMixin);
				self:Debug("          Found ButtonMixin");
				if(type(Module.ButtonMixin.Activate) == "function") then
					self:Debug("               Found Activator");			
					table.insert(FlexBar2.ButtonActivatorList, Module.ButtonMixin.Activate);
				end
				if(type(Module.ButtonMixin.Deactivate) == "function") then
					self:Debug("               Found Deactivator");
					table.insert(FlexBar2.ButtonDeactivatorList, Module.ButtonMixin.Deactivate);
				end
				if(type(Module.ButtonMixin.LoadSettings) == "function") then
					self:Debug("               Found Setting Loader");			
					table.insert(FlexBar2.ButtonSettingLoaderList, Module.ButtonMixin.LoadSettings);
				end
				if(type(Module.ButtonMixin.SetDefaults) == "function") then
					self:Debug("               Found Default Settings");			
					table.insert(FlexBar2.ButtonDefaulters, Module.ButtonMixin.SetDefaults);
				end
			end
			if(type(Module.GroupMixin) == "table") then
				self:Debug("          Found GroupMixin: ");
				self.GroupMixinList[Name] = Module.GroupMixin;
				if(type(Module.GroupMixin.Activate) == "function") then
					self:Debug("               Found Activator");
					table.insert(self.GroupActivatorList, Module.GroupMixin.Activate);
				end
				if(type(Module.GroupMixin.Deactivate) == "function") then
					self:Debug("               Found Deactivator");
					table.insert(self.GroupDeactivatorList, Module.GroupMixin.Deactivate);
				end
			end
			if(type(Module.ButtonWaterfall) == "function") then
				self:Debug("          Found button waterfall");
				self.ButtonWaterfallList[Name] = Module.ButtonWaterfall;
			end
			if(type(Module.GroupWaterfall) == "function") then
				self:Debug("          Found group waterfall");
				self.GroupWaterfallList[Name] = Module.GroupWaterfall;
			end
			if(type(Module.GeneralWaterfall) == "function") then
				self:Debug("          Found general waterfall");
				self.GeneralWaterfallList[Name] = Module.GeneralWaterfall;
			end
		end
		
		-- Create the button superclass which will inherit from all registered ButtonMixins.
		self.ButtonSuperClass = self.OO.Class(unpack(self.ButtonMixinList));
			function self.ButtonSuperClass.prototype:init(Name, ...)
				-- Call Ace2 init stuff
				FlexBar2.ButtonSuperClass.super.prototype.init(self);
				FlexBar2:Debug("          Initializing");
				if(type(Name) ~= "string") then
					FlexBar2:Debug("               Failed, invalid name");
					return false;
				end
				-- Set the button's infotable (the table which contains its saved vars) as self.InfoTable
				self.InfoTable = FlexBar2.db.char.Buttons[Name];
				FlexBar2.Buttons[Name] = self;
				-- Store the button's name into self.Name
                self.Name = Name;
				-- Add the button to the button list
				table.insert(FlexBar2.ButtonList, Name);
			end

			-- LoadSettings superfunction which loads to all "mini-loadsettings" functions that modules have declared
			function self.ButtonSuperClass.prototype:LoadSettings()
				for _, Func in ipairs(FlexBar2.ButtonSettingLoaderList) do
					Func(self);
				end
			end

			-- SetDefaults superfunction which loads to all "mini-setdefaults" functions that modules have declared
			function self.ButtonSuperClass.prototype:SetDefaults()
				for _, Func in ipairs(FlexBar2.ButtonDefaulters) do
					Func(self);
				end
			end

		-- Create the group superclass which inherits from all registered GroupMixins
		self.GroupSuperClass = self.OO.Class(unpack(self.GroupMixinList));
			function self.GroupSuperClass.prototype:init(Name, ButtonTable, ...)
				FlexBar2:Debug("          Initializing");
				-- Call Ace2 init stuff
				FlexBar2.GroupSuperClass.super.prototype.init(self);
				
				-- Store the group into flexbar's group list
                FlexBar2.Groups[Name] = self;

				-- Store the group name internally
				self.Name = Name;

				-- Store how much members the group has
                self.Count = #ButtonTable;

				-- Declare the infotable internally for easy access later on
				self.InfoTable = FlexBar2.db.char.Groups[Name];
				

				-- Create a member list which will contain all members of the group in a 'Name = Object' relation. This is derived from the ButtonTable that has been passed
                local MemberList = {};
				self.MemberList = MemberList
                local ButtonList = FlexBar2.Buttons;
                for _, ButtonName in ipairs(ButtonTable) do
					local Button = ButtonList[ButtonName];
					-- Add the button we just found to the memberlist
					MemberList[ButtonName] = ButtonList[ButtonName];

					-- 'Tell' the button what group it's in by declaring Button.Group
                    Button.Group = self;
					-- Add a list of group members to the button in Button.GroupMemberList
					Button.GroupMemberList = MemberList;                    
                end
                
				-- Loop through all activator functions if they are declared (which is quite rare)
				for _, Func in ipairs(FlexBar2.GroupActivatorList) do
					Func(self, Name, ButtonTable, ...);
				end
			end

			function self.GroupSuperClass.prototype:Deactivate()
				for _, Func in ipairs(FlexBar2.GroupDeactivatorList) do
					Func(self, Name, ButtonTable);
				end
			end
	end

	function FlexBar2:OnEnable()
		-- Call all module :Activate(); functions
		for Key, ModuleName in ipairs(FlexBar2.ModuleList) do
			local Module = FlexBar2:GetModule(ModuleName);
			if(type(Module.Activate) == "function") then
				Module:Activate();
			end
		end
		-- Setup the GUI
		self:SetupWaterfall();

		self:Debug("Generating buttons");

		-- Loop through all buttons we're suposed to create in the saved vars (and create them ofc -.-)
		for Name, InfoTable in pairs(self.db.char.Buttons) do
			if(InfoTable.Generate == true) then
				self:Debug("     Generating " .. Name);
				self:Generate(Name);
                self:Debug("Generated");
			end
		end

		-- Loop through all groups we're suposed to create and create em
        for Name, MemberList in pairs(self.db.char.Groups) do
            FlexBar2:CreateGroup(Name, MemberList);
        end
	end
	
	-- Needed to keep track of which order modules were loaded in, as we have to activate their mixins in the order they were loaded
    FlexBar2.ModuleList = {};
	function FlexBar2:OnModuleCreated(Name, Module)
		table.insert(self.ModuleList, Name);
	end
end

do -- Section [2]: Configuration: GUI & slashcommands 
	-- Slashcommand handler: thanks to Mairelon for the inspiration on the API (though most code was written myself except for ToTable)
	local SlashCommands = {};
	FlexBar2.SlashCommands = SlashCommands;

	-- Two tables keep track of all registered commands & groups
	SlashCommands.CommandList = {};
    SlashCommands.GroupList = {};
	function SlashCommands.Dispatch(Command)
		-- Refuse to do anything at all in combat
		if(InCombatLockdown()) then FlexBar2:Print("Configuration is not possible in combat"); return; end
		-- Make sure Command is a string
		if(type(Command) ~= "string") then FlexBar2:Print("Error: Command error. Please report"); return false; end
		-- Find out what function was called and what its arguments are
		local _, _, Function, Arguments = Command:find("^(%w+)[ ]*(.*)");
		-- If the user asked for help, call the help directly and return
		if(Function == "help") then SlashCommands:Help(Arguments); return; end
		-- Make sure the asked function is registered and valid before continueing
		if(type(SlashCommands.CommandList[Function]) == "table" and type(SlashCommands.CommandList[Function].Function) == "function") then
			-- Convert the param=value format into a table
			local Params = SlashCommands:ToTable(Arguments or "");
			if(SlashCommands:Verify(Function, Params, true) == false) then return false; end
			FlexBar2:Debug(Params);
			-- Call the function
			SlashCommands.CommandList[Function].Function(Params);
		else
			-- If all fails, give the user some help
			SlashCommands:Help();
		end
	end
	
	-- Register slashcommands with AceConsole
	FlexBar2:RegisterChatCommand({"/fb", "/flexbar", "/fb2", "/flexbar2"},
	{
	    type = 'text',
	    name = "FlexBar2 Slashcommands",
	    desc = "Flexbar2 slashcommand handler",
        usage = "gui will show up the config GUI. /fb help will tell you more about how to use slashcommands.",
	    get = false,
	    set = SlashCommands.Dispatch,
	});
	
	-- Help function
	function SlashCommands:Help(Group)
        FlexBar2:Debug("Group:", Group)
		-- If a group is passed we tell the user all avail commands in that group
        if(type(Group) == "string" and Group ~= "") then
            FlexBar2:Print("Available commands: ");
            for Command, InfoTable in pairs(SlashCommands.CommandList) do
                if(Group == InfoTable.Group) then
                    if(InfoTable.Help) then
						-- Found help for the command, print it out
                        FlexBar2:Print(InfoTable.Help);
                    else
						-- No help specified for this command meh, give some default message that won't enlighten the user at all.
                        FlexBar2:Print("/FlexBar2 " .. Command .. ": no help available");
                    end
                end
		    end
		-- No group was passed, print a generic help listing all possible groups
        elseif(Group == "" or type(Group) ~= "string") then
            FlexBar2:Print("Usage: /fb help <group>.");
			FlexBar2:Print("Please Choose a group to get further help");
            for GroupName, _ in pairs(SlashCommands.GroupList) do
                FlexBar2:Print(GroupName);
            end
        end
	
		
	end

	function SlashCommands:ToTable(msg) -- nearly unmodified code from Mairelon's utility class used in FB1
		if msg == nil then return {} end
		-- pattern to return name=
		local pattern = "([%a_][%w_]*)="
		local index = 1
		local params = {}
		local firsti, lasti, capture = string.find(msg,pattern)
		-- while we have a name=, process the info after it.
		while capture do
			local varname=string.lower(capture)
			index = index+lasti
			firsti, lasti, capture = string.find(string.sub(msg, index),pattern)
			if not firsti then firsti=string.len(msg) else firsti=firsti-2 end
			local str = string.sub(msg,index, index+firsti)
			if(string.find(str, "[ ]+$")) then -- code to strip out spaces at the end
				str = string.sub(str, 1, (string.find(str, "()[ ]+$")) - 1);
			end
			-- str contain the rest of the equality ( varname = str )
			params[varname] = str;
		end
		return params;
	end

	function SlashCommands:Verify(Function, Params, Output)
		-- Sanity checks to make sure everything is set up currectly
		if(type(SlashCommands.CommandList[Function]) ~= "table"
			or type(SlashCommands.CommandList[Function].Function) ~= "function"
			or type(SlashCommands.CommandList[Function].Params) ~= "table") then
			if(Output) then FlexBar2:Print("Slashcommand not set up correctly, cancelling"); end
			return false;
		end

		-- If a group has been set in the slashcommand we convert it to a set of buttons, this makes button and group interchangable
        if(type(Params["button"]) ~= "string" and type(Params["group"]) == "string") then
            Params.button = FlexBar2:GroupNameToButtonRange(Params["group"]);
        end

		-- Loop through every param that has to be verified
		for Param, Info in pairs(SlashCommands.CommandList[Function].Params) do

			-- Make sure the param is set
			if(type(Params[Param]) ~= "string") then
				-- Param is not set but is required, failure
				if(Info.Required == true) then
					if(Output) then FlexBar2:Print("Required parameter [" .. Param .. "] not set, cancelling"); end
					return false;
				-- Param not set, not required and has no default value available, pass
				elseif(type(Info.DefaultValue) ~= "string") then
					break;
				end
			end

			-- Set the default value if the param is not set and a default val has been declared
			if(type(Info.DefaultValue) == "string" and type(Params[Param]) ~= "string") then
				Params[Param] = Info.DefaultValue;
			end

			-- Make sure the param is in the table if allowed values, if it has been set
			if(type(Info.AllowedValues) == "table") then
				if(not FlexBar2:ValueIsInTable(Info.AllowedValues, Params[Param])) then
					if(Output) then FlexBar2:Print("[" .. Params[Param] .. "] is not in the list of allowed values for " .. Param); end
					return false;
				end
			end

			-- If a verify func has been set, make sure it returns true for the param, if it doesn't, then we outpit its return value
            if(type(Info.VerifyFunc) == "function") then
                local VerifyReturn = Info.VerifyFunc(Params[Param]);
                if(VerifyReturn ~= true) then
				    if(Output) then FlexBar2:Print("Verification failed for [" .. Param .. "]: " .. (VerifyReturn or ("], Invalid value [" .. Params[Param] .. "]"))); end
				    return false;
			    end
            end             
		end
		return true;
	end

	--[[
		Command: The name of the command we are adding
		Function: The command that this will be passed to
		Group: What group the function is in (for the help box)
		Help: function usage which will be displayed in the /FlexBar2 help <command> message
	--]]
	function SlashCommands:AddCommand(Command, Function, Group, Help)
		if(type(Command) ~= "string" or type(Function) ~= "function" or SlashCommands.CommandList[Command] == "table") then
			return false;
		else
			Command = Command:lower();
		end
		if(type(Group) ~= "string") then Group = "main"; end
        SlashCommands.GroupList[Group] = true;
		SlashCommands.CommandList[Command] = {Function = Function, Group = Group, Help = Help, Params = {}};
	end
	
	--[[
		Command: The name of the command we are adding parameters to
		Param: The name of the parameter we are adding
		Required: wether to error out if the parameter is not set
		Types: what type the parameter can be (string or table [ranges are seen as tables])
		AllowedValues: what values are allowed for this parameter (example: {"true", "false"})
		Default: what default value the parameter takes if it is ommited
	--]]
	function SlashCommands:AddParam(Command, Param, Required, AllowedValues, DefaultValue, VerifyFunc) 
		-- Sanity check
		if(type(Command) ~= "string"
			or type(Param) ~= "string"
            or type(SlashCommands.CommandList[Command]) ~= "table"
			or type(SlashCommands.CommandList[Command].Params[Param]) == "table") then
				return false;
		end

		-- Add the command to the list
		SlashCommands.CommandList[Command].Params[Param] =
		{
			Required = Required,
			AllowedValues = AllowedValues,
			DefaultValue = DefaultValue,
			VerifyFunc = VerifyFunc,
		};
	end

	function SlashCommands:SetAlias(Command, Alias) 
		-- Sanity check
		if(type(Command) ~= "string" or type(Alias) ~= "string" or type(SlashCommands.CommandList[Command]) ~= "table" or type(SlashCommands.CommandList[Alias]) == "table") then return; end
		SlashCommands.CommandList[Alias] = SlashCommands.CommandList[Command];
	end
	
	-- Slashcommand utility function to make sure buttons exist
    function SlashCommands.VerifyExistance(Button)
        if(not FlexBar2:ButtonsExist(Button)) then
            return "One or more buttons do not exist";
        else
            return true;
        end
    end
	
	-- Slashcommand utility function to make sure buttons do NOT exist
    function SlashCommands.VerifyNonExistance(Button)
        if(FlexBar2:ButtonsExist(Button)) then
            return "One or more buttons already exist";
        else
            return true;
        end
    end

	-- Slashcommand utility function to make sure buttons are NOT grouped
    function SlashCommands.VerifyNotGrouped(Button)
        local Existance = SlashCommands.VerifyExistance(Button);
        if(Existance ~= true) then
            return Existance;
        else
            if(not FlexBar2:ButtonIsGrouped(Button)) then
                return true;
            else
                return "One or more buttons are already grouped";
            end
        end
    end
    
	-- Slashcommand utility function to make sure buttons are grouped
    function SlashCommands.VerifyGrouped(Button)
        local Existance = SlashCommands.VerifyExistance(Button);
        if(Existance ~= true) then
            return Existance;
        else
            if(FlexBar2:ButtonIsGrouped(Button)) then
                return true;
            else
                return "One or more buttons are not grouped";
            end
        end
    end
    
	-- /fb create handler
	local function SlashCommand_Create(Params)
		for _, ButtonName in ipairs(FlexBar2:ToButtonTable(Params.button)) do
			FlexBar2:Create(ButtonName);
		end
	end

	-- /fb delete handler
	local function SlashCommand_Delete(Params)
		for _, ButtonName in ipairs(FlexBar2:ToButtonTable(Params.button)) do
			FlexBar2:Delete(ButtonName);
		end
	end

	-- /fb list handler
    local function SlashCommand_List()
        FlexBar2:Print("List of buttons: ");
        for ButtonName, _ in pairs(FlexBar2.Buttons) do
            FlexBar2:Print("[" .. ButtonName .. "]");
        end
    end
	
	-- /fb group handler
    local function SlashCommand_Group(Params)
        FlexBar2:CreateGroup(Params.name, FlexBar2:ToButtonTable(Params.button));
    end

	-- /fb ungroup handler
    local function SlashCommand_Ungroup(Params)
        local ButtonTable = FlexBar2:ToButtonTable(Params.button);
        for _, ButtonName in ipairs(ButtonTable) do
            FlexBar2:RemoveButtonFromGroup(ButtonName);
        end
    end

	-- Localize those handy utility funcs
    local VerifyExistance = SlashCommands.VerifyExistance;
    local VerifyNonExistance = SlashCommands.VerifyNonExistance;
    local VerifyGrouped = SlashCommands.VerifyGrouped;
    local VerifyNotGrouped = SlashCommands.VerifyNotGrouped;
	-- Register /fb gui command
	SlashCommands:AddCommand("gui", function() FlexBar2.Waterfall:Open("FlexBar2"); end, "main", "/FlexBar2 gui");
	-- Register /fb create command
	SlashCommands:AddCommand("create", SlashCommand_Create, "main", "/FlexBar2 create button=<Buttons>");
		SlashCommands:AddParam("create", "button", true, nil, nil, VerifyNonExistance);
	-- Register /fb delete command
	SlashCommands:AddCommand("delete", SlashCommand_Delete, "main", "/FlexBar2 delete button=<Buttons>");
	SlashCommands:SetAlias("delete", "remove");
		SlashCommands:AddParam("delete", "button", true, nil, nil, VerifyExistance);
	-- Register /fb list command
    SlashCommands:AddCommand("list", SlashCommand_List, "main", "/FlexBar2 list");
	-- Register /fb group command
    SlashCommands:AddCommand("group", SlashCommand_Group, "main", "/FlexBar2 group button=<Buttons> name=<Group name>");
        SlashCommands:AddParam("group", "button", true, nil, nil, VerifyNotGrouped);
        SlashCommands:AddParam("group", "name", true, nil, nil, function(Name) if(string.find(Name, "^[%w_]+$")) then return true; else return "Invalid group name"; end end);
	-- Register /fb ungroup command
    SlashCommands:AddCommand("ungroup", SlashCommand_Ungroup, "main", "/FlexBar2 ungroup group=<Group>");
		SlashCommands:AddParam("ungroup", "button", true, nil, nil, VerifyGrouped);
	
	-- Gui config
	function FlexBar2:SetupWaterfall() -- Called in OnEnable
		local Waterfall = self.Waterfall;
		self:Debug("Setting up Waterfall");
		-- Create a basic tree for waterfall, this is the menu on the left
		local WaterfallTree = 
		{
			{
				text = "General Settings",
				id = "gnrlset",
				isOpen = true,
			},
			{
				text = "Button Control",
				id = "buttonctrl",
				isOpen = false,
			},
			{
				text = "Group Control",
				id = "groupctrl",
				isOpen = false,
			},
		};

		-- Localize all stuff we will need in the gui here
		local buttonctrl_CreateTemp, buttonctrl_EditSave, buttonctrl_EditTemp;
		local ButtonTable, ButtonCheckOverride, OnEnd;
        local groupctrl_EditSetTemp, groupctrl_EditSetSave, groupctrl_CreateGroupNameTemp, groupctrl_CreateButtonsTemp, groupctrl_RemoveButtonTemp, groupctrl_DisbandGroupTemp;
		local function BuildWaterfallTree(Id) -- This function is called each time the ui is rebuilt
			FlexBar2:Debug("GUI Id: ", Id);
			if not Id then return ""; end
			if(Id == "gnrlset") then -- general settings
				for Module, Data in pairs(FlexBar2.GeneralWaterfallList) do
					if(type(Data) == "function") then
						Data();
					end
				end

				Waterfall:AddControl(
				"type", "button",
				"text","Cancel changes",
				"fullRefresh", true,
				"width", 130,
				"execFunc", function() --[[lol]] end);

				-- Save Button --
				Waterfall:AddControl(
				"type", "button",
				"text","Save changes",
				"fullRefresh", true,
				"noNewLine", true,
				"width", 130,
				"execFunc", function()
					--[[lol]]--
				end);
			elseif(Id == "buttonctrl") then -- Button control
                FlexBar2:Debug("editsave: ", buttonctrl_EditSave);

				-- Generate a buttontable out of the button range given by the user if he has given one. Also only do this if the user does NOT come out of the group control tab (in which case the override local will be true)
				if(buttonctrl_EditSave ~= nil and buttonctrl_EditSave ~= "" and ButtonCheckOverride ~= true) then
					ButtonTable = FlexBar2:ToButtonTable(buttonctrl_EditSave);
					FlexBar2:Debug("ButtonTable: ", #ButtonTable);
					-- Make sure buttontable is actualy valid
					if(ButtonTable == nil or #ButtonTable == "0") then buttonctrl_EditSave = nil; ButtonTable = nil; end
					-- BUttontable is valid, now lets find out if every button that the user gave actualy exists
					for _, ButtonName in ipairs(ButtonTable) do
						if(type(FlexBar2.Buttons[ButtonName]) ~= "table") then
							FlexBar2:Print("non-existant button(s) have been inputted");
							buttonctrl_EditSave = nil;
                            ButtonTable = nil;
						end
					end
				end
				-- Main buttoncontrol menu shown if no buttons have been chosen for editing
				if(type(ButtonTable) ~= "table") then
					FlexBar2:Debug("Waterfall: ButtonCtrl main");
					-- Create Button --
					Waterfall:AddControl(
					"type", "label",
					"text", "Create Button: ");
					
					Waterfall:AddControl(
					"type","textbox",
					"width",200,
					"fullRefresh", true,
					"noNewLine", true,
					"setFunc", function()
                        local ButtonTable = FlexBar2:ToButtonTable(buttonctrl_CreateTemp);
                            if(type(ButtonTable) == "table") then
                            buttonctrl_EditSave = buttonctrl_CreateTemp;
                            for _, ButtonName in ipairs(ButtonTable) do
                                FlexBar2:Create(ButtonName);
                            end
                        end
                    end,
					"getFunc", function() return ""; end,
					"changedFunc", function(ButtonName) buttonctrl_CreateTemp = ButtonName; end);
					Waterfall:AddControl(
					"type", "button",
					"text","Create",
					"fullRefresh", true,
					"noNewLine", true,
					"width", 130,
					"execFunc", function()
                        local ButtonTable = FlexBar2:ToButtonTable(buttonctrl_CreateTemp);
                            if(type(ButtonTable) == "table") then
                            buttonctrl_EditSave = buttonctrl_CreateTemp;
                            for _, ButtonName in ipairs(ButtonTable) do
                                FlexBar2:Create(ButtonName);
                            end
                        end
                    end);
					
					-- Edit Button --
					Waterfall:AddControl(
					"type", "label",
					"text", "Edit Button: ");
					
					Waterfall:AddControl(
					"type","textbox",
					"width",200,
					"noNewLine", true,
					"fullRefresh", true,
					"setFunc", function(ButtonName) buttonctrl_EditSave = ButtonName end,
					"changedFunc", function(ButtonName) buttonctrl_EditTemp = ButtonName end);
					
					Waterfall:AddControl(
					"type","button",
					"text","Edit",
					"noNewLine",true,
					"fullRefresh", true,
					"width",80,
					"execFunc", function() buttonctrl_EditSave = buttonctrl_EditTemp end);
					return "Button control";
				-- Submenu of buttoncontrol; edit part, shown if buttons have been chosen for editing
				else
					FlexBar2:Debug("Waterfall: Buttonctrl edit");
					for Module, Data in pairs(FlexBar2.ButtonWaterfallList) do
						if(type(Data) == "function") then
							Data("edit", buttonctrl_EditSave, ButtonTable);
						end
					end
					-- Delete Button --
					Waterfall:AddControl(
					"type", "button",
					"text","Delete Button",
					"fullRefresh", true,
					"width", 100,
					"execFunc", function()
                        for _, ButtonName in ipairs(ButtonTable) do
                            FlexBar2:Delete(ButtonName);
                        end
                        buttonctrl_EditSave = nil;
                        ButtonTable = nil;
                        if(ButtonCheckOverride == true) then
							if(OnEnd ~= false) then
								Waterfall:Open("FlexBar2", OnEnd);
							else
								Waterfall:Close("FlexBar2");
							end
                            ButtonCheckOverride = nil;
                        end
                        end);
					
					-- Cancel Button --
					Waterfall:AddControl(
					"type", "button",
					"text","Cancel changes",
					"fullRefresh", true,
					"noNewLine", true,
					"width", 130,
					"execFunc", function()
						buttonctrl_EditSave = nil;
                        ButtonTable = nil;
                        if(ButtonCheckOverride == true) then
							if(OnEnd ~= false) then
								Waterfall:Open("FlexBar2", OnEnd);
							else
								Waterfall:Close("FlexBar2");
							end
                            ButtonCheckOverride = nil;
                        end
						end,
					"execArg1");

					-- Save Button --
					Waterfall:AddControl(
					"type", "button",
					"text","Save changes",
					"fullRefresh", true,
					"noNewLine", true,
					"width", 130,
					"execFunc", function()
						for LoopNumber, ButtonName in ipairs(ButtonTable) do
                            for Module, Data in pairs(FlexBar2.ButtonWaterfallList) do
                                if(type(Data) == "function") then
                                    Data("save", LoopNumber, ButtonName);
                                end
                            end
                        end
						
						buttonctrl_EditSave = nil;
                        ButtonTable = nil;
                        if(ButtonCheckOverride == true) then
							if(OnEnd ~= false) then
								Waterfall:Open("FlexBar2", OnEnd);
							else
								Waterfall:Close("FlexBar2");
							end
                            ButtonCheckOverride = nil;
                        end
					end);

					return "Edit button settings";
				end
			-- Group control part
			elseif(Id == "groupctrl") then
				-- If no group has been set, show group control main menu
                if((type(groupctrl_EditSetSave) ~= "string" or type(FlexBar2.Groups[groupctrl_EditSetSave]) ~= "table")) then
					-- Handy little CreateGroup function, used so often I rather localize it then redeclare it each time
                    local CreateGroup = function()
                        ButtonTable = FlexBar2:ToButtonTable(groupctrl_CreateButtonsTemp);
                        local GroupName = groupctrl_CreateGroupNameTemp;
                        if(type(ButtonTable) == "table" and string.find(GroupName, "^[%w_]+$")) then
                            FlexBar2:CreateGroup(GroupName, ButtonTable);
                            ButtonCheckOverride = true; 
							OnEnd = "groupctrl";
                            Waterfall:Open("FlexBar2", "buttonctrl");
                        else
                            FlexBar2:Print("Invalid Group name and/or Button range");
                        end
                    end
                    
					-- Create Group --
                    Waterfall:AddControl(
                    "type", "label",
                    "text", "Create Group: ");

                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", CreateGroup,
                    "changedFunc", function(GroupName) groupctrl_CreateGroupNameTemp = GroupName end);

                    Waterfall:AddControl(
                    "type", "label",
                    "text", "                         Buttons: ");

                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", CreateGroup,
                    "changedFunc", function(ButtonRange) groupctrl_CreateButtonsTemp = ButtonRange end);
                    
                    Waterfall:AddControl(
                    "type","button",
                    "text","Create",
                    "noNewLine",true,
                    "fullRefresh", true,
                    "width",80,
                    "execFunc", CreateGroup);
					
					-- Disband Group --
                    Waterfall:AddControl(
                    "type", "label",
                    "text", "Disband group: ");

                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", function()
                        FlexBar2:DisbandGroup(groupctrl_DisbandGroupTemp);
                    end,
                    "changedFunc", function(GroupName) groupctrl_DisbandGroupTemp = GroupName end);

                    Waterfall:AddControl(
                    "type","button",
                    "text","Disband",
                    "noNewLine",true,
                    "fullRefresh", true,
                    "width",80,
                    "execFunc", function()
                        FlexBar2:DisbandGroup(groupctrl_DisbandGroupTemp);
                    end);
					
					-- Remove buttons from group (ungroup) --
                    Waterfall:AddControl(
                    "type", "label",
                    "text", "Ungroup button(s): ");

                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", function()
                        local ButtonTable = FlexBar2:ToButtonTable(groupctrl_RemoveButtonTemp);
                        if(type(ButtonTable) == "table") then
                            for _, ButtonName in ipairs(ButtonTable) do
                                FlexBar2:RemoveButtonFromGroup(ButtonName);
                            end
                        end
                    end,
                    "changedFunc", function(ButtonRange) groupctrl_RemoveButtonTemp = ButtonRange end);
                    
                    Waterfall:AddControl(
                    "type","button",
                    "text","Remove",
                    "noNewLine",true,
                    "fullRefresh", true,
                    "width",80,
                    "execFunc", function()
                        local ButtonTable = FlexBar2:ToButtonTable(groupctrl_RemoveButtonTemp);
                        if(type(ButtonTable) == "table") then
                            for _, ButtonName in ipairs(ButtonTable) do
                                FlexBar2:RemoveButtonFromGroup(ButtonName);
                            end
                        end
                    end);
					
					-- Edit groups --
                    Waterfall:AddControl(
                    "type", "label",
                    "text", "Edit Group settings: ");
                    
                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", function()
                        if(type(FlexBar2.Groups[groupctrl_EditSetTemp]) == "table") then
                            ButtonCheckOverride = true;
							OnEnd = "groupctrl";
                            ButtonTable = FlexBar2:ToButtonTable(FlexBar2.Groups[groupctrl_EditSetTemp].MemberList);
                            Waterfall:Open("FlexBar2", "buttonctrl");
                        end
                    end,
                    "changedFunc", function(GroupName) groupctrl_EditSetTemp = GroupName end);
                    
                    Waterfall:AddControl(
                    "type","button",
                    "text","Edit",
                    "noNewLine",true,
                    "fullRefresh", true,
                    "width",80,
                    "execFunc", function()
                        if(type(FlexBar2.Groups[groupctrl_EditSetTemp]) == "table") then
                            ButtonCheckOverride = true;
							OnEnd = "groupctrl";
                            ButtonTable = FlexBar2:ToButtonTable(FlexBar2.Groups[groupctrl_EditSetTemp].MemberList);
                            Waterfall:Open("FlexBar2", "buttonctrl");
                        end
                    end);
                    return "Group control";
                end
			end
		end
		
		function FlexBar2:PopupButtonGUI(Buttons)
			ButtonTable = Buttons;
			ButtonCheckOverride = true;
			OnEnd = false;
			Waterfall:Open("FlexBar2", "buttonctrl");
		end
		-- Register our config table & function with waterfall
		Waterfall:Register('FlexBar2',
		'title', 'FlexBar2 Config',
		'defaultPane', 'gnrlset',
		'width', 800,
		'height', 500,
		'tree', WaterfallTree,
		'children', BuildWaterfallTree);
	end
end

do -- Section [3]: Button control 
	-- Create a button and make sure its recreated at next login too
	function FlexBar2:Create(Name)
		-- Make sure the button doesn't already exist
		if(type(FlexBar2.db.char.Buttons[Name]) ~= "table") then
			FlexBar2.db.char.Buttons[Name] = { Generate = true};
			FlexBar2:Generate(Name);
			return true;
		else
			FlexBar2:Print("Button already exists");
			return false;
		end
	end
	
	-- Delete a button, run whatever has been set in the deactivation mixin, remove the button from its group and make sure it never comes back :)
	function FlexBar2:Delete(Name)
		-- only do something if the button actualy exists
		if(type(FlexBar2.db.char.Buttons[Name]) == "table") then
			FlexBar2.db.char.Buttons[Name] = nil;
			local Button = FlexBar2.Buttons[Name];
			-- Run all deactivators.
			for _, Func in ipairs(FlexBar2.ButtonDeactivatorList) do
				Func(Button);
			end
			-- Remove the button from any groups its in
            FlexBar2:RemoveButtonFromGroup(Name);
			-- nil the button's entry in the listing
			FlexBar2.Buttons[Name] = nil;
			return true;
		else
			-- Someone is trying to delete a button that doesnt exist, tell him
			FlexBar2:Print("Cannot delete non-existant button");
			return false;
		end
		
	end

	-- Generate a button, load its modules and wipe their settings if needed
	function FlexBar2:Generate(Name)
		-- Make sure to only do something if the button doesnt exist
		if(self.Buttons[Name] ~= nil) then
			self:Debug("          Error: Button already exists, aborting");
			return false;
		end
		-- Initialize an instance of the superclass
		self.ButtonSuperClass:new(Name);
        local ButtonList = FlexBar2.Buttons;
        if(type(self.db.char.Buttons[Name].Modules) ~= "table") then self.db.char.Buttons[Name].Modules = {}; end
        local ModuleList = self.db.char.Buttons[Name].Modules;
		-- Loop through all modules
        for _, ModuleName in ipairs(FlexBar2.ModuleList) do
            local Module = FlexBar2:GetModule(ModuleName);
            local ModuleVersion = Module.Version;
            local ButtonMixin = Module.ButtonMixin;
			-- If the module has a buttonmixin continue
            if(type(ButtonMixin) == "table") then
				-- find out if the Initialize function exists, if it does, call it
                if(type(ButtonMixin.Initialize) == "function") then ButtonMixin.Initialize(ButtonList[Name]); end
				-- If the module has never been previously loaded yet OR if the ModuleVersion integrer has been changed, call SetDefaults(), this code allows to reset the settings of one module when needed without touching the others.
                if((type(ModuleList) ~= "table" or ModuleList[ModuleName] ~= ModuleVersion) and type(ButtonMixin.SetDefaults) == "function") then
                    ButtonMixin.SetDefaults(ButtonList[Name]);
                end
				-- Call loadsettings for the button if it exists
                if(type(ButtonMixin.LoadSettings) == "function") then ButtonMixin.LoadSettings(ButtonList[Name]); end
				-- Finaly, call Activate on the button if it exists
                if(type(ButtonMixin.Activate) == "function") then ButtonMixin.Activate(ButtonList[Name]); end
				-- Save the data version of the module  in the saved vars
                ModuleList[ModuleName] = ModuleVersion;
            end
        end
        return true;
	end
end

do -- Section [4]: Group control
    function FlexBar2:CreateGroup(Name, ButtonTable)
        self.GroupSuperClass:new(Name, ButtonTable);
        self.db.char.Groups[Name] = ButtonTable;
    end
	
	function FlexBar2:DisbandGroup(Name)
		-- Make sure the grp exists first
        if(type(FlexBar2.Groups[Name]) == "table") then
            local Group = FlexBar2.Groups[Name];
			-- Remove every button from the group, fast, no need to keep the count or anything like RemoveButtonFromGroup does, we're disbanding it anyway
            for ButtonName, Button in pairs(Group.MemberList) do
                Button.Group = nil;
                Button.GroupMemberList = nil;
            end
			Group:Deactivate();
            FlexBar2.Groups[Name] = nil;
            FlexBar2.db.char.Groups[Name] = nil;
        end
    end

    function FlexBar2:AddButtonToGroup(Group, Name)
		-- Make sure the button is not already grouped and the group exists before adding;
		if(not FlexBar2:ButtonIsGrouped(Name) and type(FlexBar2.Groups[Group]) == "table") then
			-- Localize
			local Group = FlexBar2.Groups[Group];
			local Button = FlexBar2.Buttons[Name];
			-- Add the button to the list of members of the group
			Group.MemberList[Name] = Button;
			-- Increment the member count of this group by one
			Group.Count = Group.Count + 1;
			-- Declare Group and GroupMemberList on the button
			Button.Group = Group;
			Button.GroupMemberList = Group.MemberList;
			-- Save the new group member
			table.insert(Group.InfoTable, Name);
        end
	end

	function FlexBar2:RemoveButtonFromGroup(Name)
		local Button = FlexBar2.Buttons[Name];
		-- Make sure its grouped first
		if(type(Button.GroupMemberList) == "table") then
			-- nil out the button from the member list
			Button.GroupMemberList[Name] = nil;
            local Group = Button.Group;
			-- Decrement the count by one
            Group.Count = Group.Count - 1;
			-- remove the button from the saved button member list
            FlexBar2:RemoveByValue(Button.Group.InfoTable, Name);
			-- nil out Group & GroupMemberList
            Button.Group = nil;
            Button.GroupMemberList = nil;
			-- If the count reached 0, just disband the group, its useless anyway :)
			if(Group.Count == 0) then
                FlexBar2:DisbandGroup(Group.Name);
            end
		end
	end
end

do -- Section [5]: Utility
	-- Gets a specific setting of a buttontable, if there's more then one button, will take the setting of the first one. accepts either a table or a string
	function FlexBar2:GetSetting(ButtonTable, Setting)
		if(type(ButtonTable) == "table") then
			ButtonName = ButtonTable[1]
		else
			ButtonName = ButtonTable;
		end
		if(type(ButtonName) ~= "string" or type(Setting) ~= "string") then return nil; end
		return FlexBar2.db.char.Buttons[ButtonName][Setting];
	end

	-- Apply a setting to a buttontable or to a button name
	function FlexBar2:ApplySetting(ButtonTable, Setting, Value)
		if(type(ButtonTable) == "table") then
			for _, ButtonName in ipairs(ButtonTable) do
				FlexBar2.db.char.Buttons[ButtonName][Setting] = Value;
			end
			return true;
		elseif(type(ButtonTable) == "string") then
			FlexBar2.db.char.Buttons[ButtonTable][Setting] = Value;
			return true;
		else
			return false;
		end
	end
	
	-- Get a buttontable out of a range of buttons, mainly used for configuration
	function FlexBar2:ToButtonTable(ButtonRange)
		-- If the given range is a string, try to match it
		local ButtonTable;
		if(type(ButtonRange) == "string") then
			ButtonTable = {};
			-- try to match a range (number)-(number)
            if(ButtonRange:find("%d+-%d+")) then
				local _, _, First, Last = ButtonRange:find("^(%d+)-(%d+)$");
				for ButtonNumber = First, Last do
					table.insert(ButtonTable, tostring(ButtonNumber));
				end
			-- Try to match a table (name);(name);(name);...
			elseif(ButtonRange:find("[%w_];[%w_]")) then
				ButtonTable = { string.split(";", ButtonRange) };
			-- No match, just assume the given string is a valid button name, for now.
			else
				ButtonTable = { ButtonRange };
			end
		-- Its a table, try to convert
		elseif(type(ButtonRange) == "table") then
            ButtonTable = {};
			for Name, Object in pairs(ButtonRange) do
                table.insert(ButtonTable, Name);
            end
			-- Failed to convert, lets return the original table again in case it was already a button table (who knows)
            if(#ButtonTable == 0) then
                return ButtonRange;
            end
		-- Not a string or not a table, fail
		else
			return false;
		end
		-- Make sure the matched buttons all matched the validity rules (only digits , letters & underscores) before returning
		for _, ButtonName in ipairs(ButtonTable) do
			if(not ButtonName:find("^[%w_]+$")) then
				return false;
			end
		end
		return ButtonTable;
	end

	-- Convert the name of a group to a range of buttons, used in the slashcommand engine to substitute a group name for a list of buttons
    function FlexBar2:GroupNameToButtonRange(GroupName)
		-- Make sure the group exists
        if(type(FlexBar2.Groups[GroupName]) == "table" and type(FlexBar2.Groups[GroupName].MemberList) == "table") then
			-- Convert the member list into a valid button table
            local ButtonTable = FlexBar2:ToButtonTable(FlexBar2.Groups[GroupName].MemberList);
            if(type(ButtonTable) == "table") then
				-- Concat the button table into a string if the buttontable is valid
                return table.concat(ButtonTable, ";");
            end
        end
        return false;
    end
	
	-- Make sure a button exists, handy for slashcommands & gui config
	function FlexBar2:ButtonsExist(Button)
		if(type(Button) == "string") then
            Button = FlexBar2:ToButtonTable(Button);
		end
		if(type(Button) == "table") then
            local ButtonList = FlexBar2.Buttons;
			for _, ButtonName in ipairs(Button) do
				if(ButtonList[ButtonName] == nil) then return false; end
			end
			return true;
		else
			return false;
		end
	end
	
	-- Make sure a button is grouped, handy for slashcommands & gui config
    function FlexBar2:ButtonIsGrouped(Button)
		if(type(Button) == "string") then
			Button = FlexBar2:ToButtonTable(Button);
		end
		if(type(Button) == "table") then
            local ButtonList = FlexBar2.Buttons;
			for _, ButtonName in ipairs(Button) do
                if(type(ButtonList[ButtonName].Group) == "table") then return true; end
			end
			return false;
		else
			return false;
		end
	end
	
	-- Table manipulation functions, they should be pretty clear in what they do
	function FlexBar2:ValueIsInTable(Table, Value)
		for n, v in pairs(Table) do
            if(v == Value) then return true; end
        end
        return false;
	end

    function FlexBar2:RemoveByValue(tbl, val)
        if(type(tbl) ~= "table") then return false; end
        for n, v in pairs(tbl) do
            if(v == val) then
                table.remove(tbl, n);
                return true;
            end
        end
        return false;
    end

    function FlexBar2:GetKeyByValue(tbl, val)
        for n, v in pairs(tbl) do
            if(v == val) then
                return n;
            end
        end
        return false;
    end
end
